//
// adi_ttcp_main.c
//
// Contains the main control loop of ADI-TTCP and implements the RELAY command.
//
// It opens a socket and listens for a controller to connect. It reads and
// actions commands and returns responses. The main loop only returns when
// it receives a QUIT command. Otherwise, when a controller closes the control
// stream it goes back to listening for another controller to connect.
//
#include "config.h"
#include "adi_ttcp.h"
#include "getopt.h"

// forward declarations for functions
static int ReadCommand(char* buf, int buf_len);
static void CloseRelays(void);
static void DestroyTransfer(PTEST_BLOCK ptfr);
static void DestroyRelay(PRELAY_BLOCK prly);
static PTEST_BLOCK BuildTransfer(char* cmd);
static int TTCP_ParseCommandLine( PCMD_BLOCK pCmdBlk, int argc, char **argv );
static PRELAY_BLOCK BuildRelay(char* cmd);
static void TTCP_SetConfigDefaults( PCMD_BLOCK pCmdBlk );
static ULONG ReadIPaddr(PCHAR ipchar, PMSG_BLOCK pmsg);

// static data definitions
static PTEST_BLOCK tfr_list  = NULL; // head of linked list of TRANSFER requests
static PRELAY_BLOCK rly_list = NULL; // head of linked list of RELAY requests

static int next_tfr_id = 1;

//global data definitions
SOCKET control_fd;
USHORT control_port = CONTROL_PORT;

//
// TTCP_MainLoop
//
int TTCP_MainLoop()
{
    SOCKADDR_IN listen_addr, control_addr;
    SOCKET      listen_fd;
    int         control_addr_len = sizeof(control_addr);

    // create a listening socket
    if ( (listen_fd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
    {
	DWORD nError = TTCP_GetLastError();

	TTCP_ConsoleMsg(
	  "ERROR: failed to create control stream listening socket\n"
	  "       TCP/IP library result code: %d (0x%8.8X)\n",
	  nError, nError);
	return (1);
    }

    // bind to listening port
    listen_addr.sin_family = AF_INET;
    listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    listen_addr.sin_port = htons(control_port);
    if ( bind(listen_fd, (SOCKADDR*)&listen_addr, sizeof(listen_addr))
	    == SOCKET_ERROR )
    {
	DWORD nError = TTCP_GetLastError();

	TTCP_ConsoleMsg(
	  "ERROR: failed to bind control stream listening socket\n"
	  "       TCP/IP library result code: %d (0x%8.8X)\n",
	  nError, nError);
	return (2);
    }

    // enter listening state, with no queue
    listen(listen_fd, 0);

    // accept sequential control connections
    while (1)
    {
	char buf[32];

	// clear out any un-run TRANSFERs
	while (tfr_list)
	{
	    PTEST_BLOCK ptfr = tfr_list;

	    tfr_list = tfr_list->m_chain;
	    DestroyTransfer(ptfr);
	}

	TTCP_ConsoleMsg("Listening for control connection on port %d\n",
	                control_port);

	if ( (control_fd =
	        accept(listen_fd, (SOCKADDR*)&control_addr, &control_addr_len))
	          == INVALID_SOCKET )
	{
	    DWORD nError = TTCP_GetLastError();

	    TTCP_ConsoleMsg(
	      "ERROR: failed to accept control stream connection\n"
	      "       TCP/IP library result code: %d (0x%8.8X)\n",
	      nError, nError);
	    return (3);
	}

	TTCP_ConsoleMsg("Control connection accepted from %s\n",
	                TTCP_GetAddr(&control_addr, buf));

	// loop reading commands and returning responses
	while (1)
	{
	    static char buf[128];
	    int n = ReadCommand(buf, sizeof(buf));

	    if ( n < 0 )
	    {
		// failed to read control stream (already reported)
		// just exit
		return (4);
	    }

	    if ( n == 0 )
	    {
		// controller closed stream; close socket
		closesocket(control_fd);

		CloseRelays();

		// terminate reading loop and accept another control stream
		break;
	    }

	    TTCP_ConsoleMsg("%s", buf);

	    if ( buf[n-1] == '\n' )
		buf[n-1] = '\0', n -= 1;
	    else
		buf[n] = '\0';

	    if ( !strncmp(buf, "QUIT", 4) )
	    {
		// close all sockets and return to caller
		closesocket(control_fd);
		closesocket(listen_fd);
		CloseRelays();
		return (0);
	    }

	    if ( !strncmp(buf, "TRANSFER", 8) )
	    {
		PTEST_BLOCK ptfr;

		ptfr = BuildTransfer(buf + 8);

		if ( ptfr && !ptfr->m_Cmd.m_msgBlk.m_msgs)
		{
		    // send back transfer identifier as response
		    char id[8];

		    sprintf(id, "%d\r\n", ptfr->m_id);
		    send(control_fd, id, (int)strlen(id), 0);

		    // add transfer to list
		    ptfr->m_chain = tfr_list;
		    tfr_list = ptfr;
		}
		else if ( ptfr )
		{
		    // send back error messages and free transfer block
		    char* p = ptfr->m_Cmd.m_msgBlk.m_msgs;

		    while (*p)
		    {
			char* q = strchr(p, '\n');

			sprintf(buf, "ERROR: %.*s", q-p+1, p);
			send(control_fd, buf, (int)strlen(buf), 0);
			p = q + 1;
		    }
		    DestroyTransfer(ptfr);
		}
		else
		{
		    char* err = "ERROR: could not create TRANSFER block\r\n";

		    send(control_fd, err, (int)strlen(err), 0);
		}
	    }
	    else if ( !strncmp(buf, "RELAY", 5) )
	    {
		PRELAY_BLOCK prly;

		prly = BuildRelay(buf + 5);

		if ( prly && !prly->m_msgBlk.m_msgs )
		{
		    // send the relay listen port as response
		    char port[12];

		    sprintf(port, "%hu\r\n", prly->m_listen_port);
		    send(control_fd, port, (int)strlen(port), 0);

		    // add relay to list
		    prly->m_chain = rly_list;
		    rly_list = prly;
		}
		else if ( prly )
		{
		    // send back error messages and free relay block
		    char* p = prly->m_msgBlk.m_msgs;

		    while (*p)
		    {
			char* q = strchr(p, '\n');

			sprintf(buf, "%.*s", q-p+1, p);
			send(control_fd, buf, (int)strlen(buf), 0);
			p = q + 1;
		    }
		    DestroyRelay(prly);
		}
		else
		{
		    char* err = "ERROR: could not create RELAY block\r\n";

		    send(control_fd, err, (int)strlen(err), 0);
		}
	    }
	    else if ( !strncmp(buf, "RUN", 3) )
	    {
		if ( tfr_list == NULL )
		{
		    char* err = "ERROR: no transfers to run\r\n";

		    send(control_fd, err, (int)strlen(err), 0);
		    continue;
		}

		TTCP_RunTransfers(tfr_list, TTCP_DispatchTransfer);

		// send back each transfer's results and free transfer block
		while ( tfr_list != NULL )
		{
		    PTEST_BLOCK ptfr = tfr_list;

		    tfr_list = tfr_list->m_chain;

		    sprintf(buf, "%d START RESULTS\r\n", ptfr->m_id);
		    send(control_fd, buf, (int)strlen(buf), 0);
		    send(control_fd,
		         ptfr->m_Cmd.m_msgBlk.m_msgs,
		         (int)strlen(ptfr->m_Cmd.m_msgBlk.m_msgs),
		         0);
		    sprintf(buf, "%d END RESULTS\r\n", ptfr->m_id);
		    send(control_fd, buf, (int)strlen(buf), 0);

		    DestroyTransfer(ptfr);
		}
	    }
	    else
	    {
		char* err = "ERROR: command not recognised\r\n";

		TTCP_ConsoleMsg(err);
		send(control_fd, err, (int)strlen(err), 0);
	    }
	}
    }

}


// ReadCommand()
//
// Receive data from control stream and assemble into a command line in
// supplied buffer. Return size of command line, including terminating '\n'.
//
static int ReadCommand(char* buf, int buf_len)
{
    int n;
    int t = 0;
    char* p = buf;

    while (1)
    {
	int max_len = buf_len - 1 - t;

	if ( max_len <= 0 )
	{
	    TTCP_ConsoleMsg("ERROR: command too long for buffer\n"
	                    "%.25s....\n", buf);
	    return (-1);
	}

	if ( (n = recv(control_fd, p, max_len, 0)) == SOCKET_ERROR )
	{
	    DWORD nError = TTCP_GetLastError();

	    TTCP_ConsoleMsg(
		"ERROR: failed to read control command\n"
		"       TCP/IP library result code: %d (0x%8.8X)\n",
		nError, nError);
	    return (-1);
	}

	if ( n == 0 ) return (0); // control stream closed

	p += n;
	t += n;

	if ( *(p-1) == '\n' )
	{
	    // end of command found - return whole command length
	    *p = '\0';
	    return (t);
	}

	// go round again to get next instalment of command
    }
}

static void CloseRelays(void)
{
    while (rly_list)
    {
	PRELAY_BLOCK prly = rly_list;

	rly_list = rly_list->m_chain;
	DestroyRelay(prly);
    }
}

static void DestroyTransfer(PTEST_BLOCK ptfr)
{
    if (!ptfr)
	return;

    if (ptfr->m_Cmd.m_msgBlk.m_msgs)
	free(ptfr->m_Cmd.m_msgBlk.m_msgs);

    if (ptfr->m_pBufBase)
	free(ptfr->m_pBufBase);

    if (ptfr->m_Socket_fd != INVALID_SOCKET)
	closesocket(ptfr->m_Socket_fd);

    free(ptfr);
}

static void DestroyRelay(PRELAY_BLOCK prly)
{
    if (!prly)
	return;

    if (prly->m_msgBlk.m_msgs)
	free(prly->m_msgBlk.m_msgs);

    if (prly->m_remote_socket != INVALID_SOCKET)
	closesocket(prly->m_remote_socket);

    if (prly->m_listen_socket != INVALID_SOCKET)
	closesocket(prly->m_listen_socket);

    free(prly);
}

static PTEST_BLOCK BuildTransfer(char* cmd)
{
    PTEST_BLOCK ptfr;
    char* argv[32];
    int argc;

    // calloc() a TEST_BLOCK and initialise the CMD_BLOCK portion
    ptfr = calloc(1, sizeof(TEST_BLOCK));
    if ( ptfr == NULL)
	return (NULL);
    TTCP_SetConfigDefaults(&ptfr->m_Cmd);

    // build argv strings from the TRANSFER line contents
    argv[0] = "ADI-TTCP";
    argc = 1;
    while (*cmd)
    {
	while (isspace(*cmd))
	    cmd += 1;
	if (!*cmd)
	    break;
	argv[argc++] = cmd;
	while (*cmd && !isspace(*cmd))
	    cmd += 1;
	if (!*cmd)
	    break;
	*cmd++ = '\0';
    }
    argv[argc] = NULL;

    // parse the TRANSFER command line arguments
    if (TTCP_ParseCommandLine(&ptfr->m_Cmd, argc, argv) == 0)
    {
	// ensure that a port address has been supplied
	if (ptfr->m_Cmd.m_Port == PORT_UNSET)
	    TTCP_AddMsg(&ptfr->m_Cmd.m_msgBlk,
	                "%s", "no port number supplied\r\n");

	ptfr->m_Cmd.m_RemoteAddr.sin_port = htons( ptfr->m_Cmd.m_Port );

	// if transmit, there had better be a target IP address
	if (ptfr->m_Cmd.m_bTransmit == TRUE)
	{
	    if (optind < argc)
	    {
		ptfr->m_Cmd.m_RemoteAddr.sin_family = AF_INET;
		ptfr->m_Cmd.m_RemoteAddr.sin_addr.s_addr =
			    ReadIPaddr(argv[optind], &ptfr->m_Cmd.m_msgBlk);
	    }
	    else
		TTCP_AddMsg(&ptfr->m_Cmd.m_msgBlk,
	                    "%s", "no target IP address supplied\r\n");
	}
    }

    // assign a transfer identifier if everything is OK
    if (ptfr->m_Cmd.m_msgBlk.m_msgs == NULL)
	ptfr->m_id = next_tfr_id++;

    return (ptfr);
}

static void TTCP_SetConfigDefaults( PCMD_BLOCK pCmdBlk )
{
    pCmdBlk->m_bTransmit          = FALSE;
    pCmdBlk->m_Protocol           = IPPROTO_TCP;
    pCmdBlk->m_Port               = PORT_UNSET;
    pCmdBlk->m_bOptContinuous     = FALSE;
    pCmdBlk->m_nOptWriteDelay     = 0;
    pCmdBlk->m_bTouchRecvData     = FALSE;
    pCmdBlk->m_nNumBuffersToSend  = 20000;
    pCmdBlk->m_nBufferSize        = 1000;
    pCmdBlk->m_nBufOffset         = 0;
    pCmdBlk->m_nBufAlign          = 0;
    pCmdBlk->m_bUseSockOptBufSize = FALSE;
    pCmdBlk->m_nSockOptBufSize    = 0;
    pCmdBlk->m_nSockOptNoDelay    = 0;
    pCmdBlk->m_bSockOptDebug      = FALSE;
    pCmdBlk->m_TokenRate          = 0;
    pCmdBlk->m_fmt                = 'K';
    pCmdBlk->m_ExtraSocket        = INVALID_SOCKET;
}

/////////////////////////////////////////////////////////////////////////////
//// TTCP_ParseCommandLine
//
// Purpose
// Parse the TRANSFER command arguments and setup the CMD_BLOCK.
//
// Parameters
//   pCmdBlk - Pointer to CMD_BLOCK to be filled with parsed option
//             information.
//   argc,   - Count and pointers to space-delimited items on
//   argv      TRANSFER command line.
//
// Return Value
// Returns zero(0) for normal return. Non-zero otherwise.
//
static int TTCP_ParseCommandLine( PCMD_BLOCK pCmdBlk, int argc, char **argv )
{
    int   c;

    if (argc < 2)
    {
        TTCP_AddMsg(&pCmdBlk->m_msgBlk, "%s", "too few arguments\r\n");
        return( !0 );
    }

    initopt();

    while (optind != argc)
    {
        c = getopt(argc, argv, "cdrtuDTb:f:l:n:p:w:A:L:O:P:" );

        switch (c)
        {
        case EOF:
            return( 0 );

        case 't':
            pCmdBlk->m_bTransmit = TRUE;
            break;

        case 'r':
            pCmdBlk->m_bTransmit = FALSE;
            break;

        case 'c':
            pCmdBlk->m_bOptContinuous = TRUE;
            break;

        case 'd':
            pCmdBlk->m_bSockOptDebug = 1;
            break;

        case 'D':
            pCmdBlk->m_nSockOptNoDelay = 1;
            break;

        case 'n':
            pCmdBlk->m_nNumBuffersToSend = atoi(optarg);
            break;

        case 'l':
            pCmdBlk->m_nBufferSize = atoi(optarg);
            break;

        case 'p':
            pCmdBlk->m_Port = atoi(optarg);
            break;

        case 'u':
            pCmdBlk->m_Protocol = IPPROTO_UDP;
            break;

        case 'A':
            pCmdBlk->m_nBufAlign = atoi(optarg);
            break;

        case 'O':
            pCmdBlk->m_nBufOffset = atoi(optarg);
            break;

        case 'b':
            pCmdBlk->m_bUseSockOptBufSize = TRUE;
            pCmdBlk->m_nSockOptBufSize = atoi(optarg);
            if( pCmdBlk->m_nSockOptBufSize < 0 )
            {
                pCmdBlk->m_nSockOptBufSize = 0;
            }
            break;

        case 'f':
            pCmdBlk->m_fmt = *optarg;
            break;

        case 'T':
            pCmdBlk->m_bTouchRecvData = 1;
            break;

        case 'w':
            pCmdBlk->m_nOptWriteDelay = atoi(optarg);
            break;

        case 'L':
            pCmdBlk->m_TokenRate = atoi(optarg);
            if( pCmdBlk->m_TokenRate < 0 )
            {
                pCmdBlk->m_TokenRate = 0;
            }
            break;

        case 'P':
            pCmdBlk->m_period = (atoi(optarg) * 1000);
            break;

        default:
            TTCP_AddMsg(&pCmdBlk->m_msgBlk,
	                "ERROR: unknown argument: '%c'\r\n", c);
            return( !0 );
        }
    }

    return( 0 );
}

static PRELAY_BLOCK BuildRelay(char* cmd)
{
    PRELAY_BLOCK prly = calloc(1, sizeof(RELAY_BLOCK));
    PMSG_BLOCK   pmsg;
    SOCKADDR_IN  local_addr,
                 remote_addr,
		 listen_addr;
    USHORT       listen_port;

    if (!prly)
	return (NULL);

    prly->m_listen_socket =
	prly->m_control_socket = prly->m_remote_socket = INVALID_SOCKET;

    pmsg = &prly->m_msgBlk;

    // extract remote machine's IP address from command line
    while (isspace(*cmd))
	cmd += 1;
    if (!*cmd)
    {
	TTCP_AddMsg(pmsg, "ERROR: no IP address supplied for remote machine\r\n");
	return (prly);
    }
    remote_addr.sin_addr.s_addr = ReadIPaddr(cmd, pmsg);

    // extract remote machine's control port number if supplied
    while (*cmd)
	cmd += 1; // pass over the IP address
    while (isspace(*cmd))
	cmd += 1; // pass over space or '\r' '\n'
    if (*cmd)
	remote_addr.sin_port = htons((short)atoi(cmd));
    else
	remote_addr.sin_port = htons((short)CONTROL_PORT);

    remote_addr.sin_family = AF_INET;

    // connect to remote machine's control stream
    if( (prly->m_remote_socket =
	    socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
    {
        TTCP_AddMsg(pmsg, "ERROR: failed to create RELAY socket (%d)\r\n",
	                  TTCP_GetLastError());
	return (prly);
    }

    local_addr.sin_family      = AF_INET;
    local_addr.sin_addr.s_addr = INADDR_ANY;
    local_addr.sin_port        = 0;
    if (bind(prly->m_remote_socket, (PSOCKADDR)&local_addr, sizeof(local_addr))
	    == SOCKET_ERROR)
    {
	TTCP_AddMsg(pmsg, "ERROR: failed to bind RELAY socket (%d)\r\n",
			  TTCP_GetLastError());
	return (prly);
    }

    if (connect(prly->m_remote_socket, (PSOCKADDR)&remote_addr, sizeof(remote_addr))
	    == SOCKET_ERROR)
    {
	TTCP_AddMsg(pmsg,
	    "ERROR: failed to connect to remote control stream (%d)\r\n",
	    TTCP_GetLastError());
	return (prly);
    }

    // create listening socket for RELAY instigator to connect to
    if ((prly->m_listen_socket = socket(AF_INET, SOCK_STREAM, 0))
	    == INVALID_SOCKET)
    {
	TTCP_AddMsg(pmsg,
	    "ERROR: failed to create RELAY listening socket (%d)\r\n",
	    TTCP_GetLastError());
	return (prly);
    }
    for (listen_port = LISTEN_PORT_BASE;
	 listen_port <= LISTEN_PORT_MAX;
         listen_port += 1)
    {
	listen_addr.sin_family = AF_INET;
	listen_addr.sin_addr.s_addr = INADDR_ANY;
	listen_addr.sin_port = htons(listen_port);
	if (bind(prly->m_listen_socket,
	         (PSOCKADDR)&listen_addr,
		 sizeof(listen_addr)) != SOCKET_ERROR)
	    break;
    }
    if (listen_port > LISTEN_PORT_MAX)
    {
	TTCP_AddMsg(pmsg, "ERROR: failed to bind RELAY listening socket (%d)\r\n",
	                  TTCP_GetLastError());
	return (prly);
    }
    listen(prly->m_listen_socket, 0);

    // create and run a thread to execute RELAY command
    prly->m_listen_port = listen_port;
    TTCP_CreateRelay(prly, TTCP_RunRelay);

    return (prly);
}

// convert an IP address in character dotted integer notation to a ULONG
static ULONG ReadIPaddr(PCHAR ipchar, PMSG_BLOCK pmsg)
{
    struct in_addr ip;

    if ((ip.s_addr = inet_addr(ipchar)) == INADDR_NONE)
	TTCP_AddMsg(pmsg, "ERROR: bad IP address: %s\r\n", ipchar);

    return (ip.s_addr);
}

//
// return character representation of IP address and port
//
char* TTCP_GetAddr(SOCKADDR_IN* sa, char* buf)
{
    sprintf(buf, "%s port %hu", INET_NTOA(sa->sin_addr), ntohs(sa->sin_port));
    return buf;
}

// format and copy a new message to list maintained by pmsg
// return pointer to new message or NULL if anything went wrong
char* TTCP_AddMsg(PMSG_BLOCK pmsg, char* fmt, ...)
{
    char buf[128];
    size_t bufLen;
    size_t msgLen;
    size_t newLen;
    PCHAR  newMsgs;
    va_list marker;

    // format the new message
    va_start(marker, fmt);
    bufLen = vsprintf(buf, fmt, marker);
    va_end(marker);

    // first message?
    if (pmsg->m_msgs == NULL)
    {
	pmsg->m_msgs = calloc(256, 1);
	if (pmsg->m_msgs == NULL)
	    return (NULL);
	pmsg->m_msgs_capacity = 255;
    }

    msgLen  = strlen(pmsg->m_msgs);

    // is there room?
    if (bufLen > pmsg->m_msgs_capacity)
    {
	// make message area bigger
	newLen  = msgLen+bufLen+64;
	newMsgs = realloc(pmsg->m_msgs, newLen);

	if (newMsgs == NULL)
	    return (NULL);
	pmsg->m_msgs = newMsgs;
	pmsg->m_msgs_capacity = newLen - msgLen;
    }

    // copy in the new message
    strcat(pmsg->m_msgs, buf);
    pmsg->m_msgs_capacity -= bufLen;

    return (pmsg->m_msgs + msgLen);
}
